/* FONT PATH
 * -------------------------- */

@font-face {
  font-family: 'FontAwesome';
  src:  url('@{fa-font-path}/fontawesome-webfont.ttf?v=@{fa-version}') format('truetype');
  font-weight: normal;
  font-style: normal;
}
