//
// For config pages in /install/html/ and /install/migration/
//

.info-page .navbar-header {
    line-height: 40px;
}

.browser-config {

    h2 {
        margin-top: 1.5em;
    }

    .disabled {
        color: #969696;
    }
}

.ssbrowser {
    h2 > img {
        height: 22px;
        vertical-align: bottom;
        margin-right: 0.5em;
    }
}