//
// Font face definitions are different than original RCUE ones.
// Here we supply only ttf variants. Additionally local variant can be
// used so it can save some load time in some browsers.
//
// This approach has a disadvantage:
// - fonts do not work in IE8 and below
//
// Advantage is that it nicely fits into Fedora packaging guidelines
//

@font-face {
    font-family: 'Open Sans';
    font-style: normal;
    font-weight: 400;
    src: local('Open Sans'), local('OpenSans'),
        url('../fonts/open-sans/OpenSans-Regular.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: normal;
    font-weight: 300;
    src:  local('OpenSans-Light'), local('Open Sans Light'),
        url('../fonts/open-sans/OpenSans-Light.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: normal;
    font-weight: 600;
    src: local('Open Sans Semibold'), local('OpenSans-Semibold'),
        url('../fonts/open-sans/OpenSans-Semibold.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: normal;
    font-weight: 700;
    src: local('Open Sans Bold'), local('OpenSans-Bold'),
        url('../fonts/open-sans/OpenSans-Bold.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: normal;
    font-weight: 800;
    src: local('Open Sans Extrabold'), local('OpenSans-Extrabold'),
        url('../fonts/open-sans/OpenSans-ExtraBold.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: italic;
    font-weight: 300;
    src: local('Open Sans Light Italic'), local('OpenSansLight-Italic'),
        url('../fonts/open-sans/OpenSans-LightItalic.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: italic;
    font-weight: 400;
    src: local('Open Sans Italic'), local('OpenSans-Italic'),
        url('../fonts/open-sans/OpenSans-Italic.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: italic;
    font-weight: 600;
    src: local('Open Sans Semibold Italic'), local('OpenSans-SemiboldItalic'),
        url('../fonts/open-sans/OpenSans-SemiboldItalic.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: italic;
    font-weight: 700;
    src: local('Open Sans Bold Italic'), local('OpenSans-BoldItalic'),
        url('../fonts/open-sans/OpenSans-BoldItalic.ttf') format('truetype');
}

@font-face {
    font-family: 'Open Sans';
    font-style: italic;
    font-weight: 800;
    src: local('Open Sans Extrabold Italic'), local('OpenSans-ExtraboldItalic'),
        url('../fonts/open-sans/OpenSans-ExtraBoldItalic.ttf') format('truetype');
}
