//
// Alerts
// --------------------------------------------------

// Add styles for Font Awesome

.alert {
    > .fa, > .fa-layered {
        font-size: 20px;
        position: absolute;
        left: 7px;
        top: 7px;
    }
    .fa-info {
        color: #72767b;
    }
}

.alert-success .fa-check-circle-o {
    color: #5cb75c;
}

.alert-info .fa-info-circle {
    color: #ccc;
}

.alert-danger .fa-exclamation-circle {
    color: #c90813;
}

.alert-warning .fa-warning {
    color: #eb7720;
}
