/* FreeIPA styles */

@import "font-awesome/variables";
@import "bootstrap/variables";
@import "patternfly/variables";
@import "variables";

@import "mixins";
@import "layout";
@import "forms-override";
@import "alerts";
@import "widgets";
@import "plugins/otp";
@import "login.less";
@import "config-pages.less";
@import "brand";
