/**
 *  Authors:
 *    UXD team
 *    Petr Vobornik <pvoborni@redhat.com>
 *
 * Copyright (C) 2013 Red Hat
 * see file 'COPYING' for use and warranty information
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.facet[name=login] {
    height: 100%;
}

.login-pf-body {
    height: 100%;
    background: @login-bg-color url("@{img-path}/@{img-bg-login}") repeat-x 50% 0;
    background-size: auto;
    color: #fff;

    .has-error .control-label {
        color: #fff;
    }
}

.reset-login-pf-height() {
    .login-pf {
        height: auto;
        .login-pf-body {
            height: auto;
        }
    }
    .login-pf .container {
        position: relative;
    }
}

// 768px = @screen-sm-min)
@media (min-width: 768px) {
    .login-pf-body {
        background-size: 100% auto;
    }
}

@media (max-height: 300px) {
    .reset-login-pf-height;
}